'use strict';

var Prism = require('prismjs');
var Slate = require('slate');
var React = require('react');

var MARK_TYPE = 'prism-token';

/**
 * Default filter for code blocks
 * @param {Node} node
 * @return {Boolean}
 */
function defaultOnlyIn(node) {
    return node.kind === 'block' && node.type === 'code_block';
}

/**
 * Default getter for syntax
 * @param {Node} node
 * @return {String}
 */
function defaultGetSyntax(node) {
    return 'javascript';
}

/**
 * Default rendering for tokens
 * @param {Object} props
 * @param {Slate.Mark} props.mark
 * @return {React.Element}
 */
function defaultTokenRender(props) {
    var mark = props.mark;

    var className = mark.data.get('className');
    return React.createElement(
        'span',
        { className: className },
        props.children
    );
}

/**
 * A Slate plugin to handle keyboard events in code blocks.
 * @return {Object}
 */

function PrismPlugin(opts) {
    opts = opts || {};
    opts.onlyIn = opts.onlyIn || defaultOnlyIn;
    opts.getSyntax = opts.getSyntax || defaultGetSyntax;
    opts.renderToken = opts.renderToken || defaultTokenRender;

    function decorate(block) {
        var grammarName = opts.getSyntax(block);
        var grammar = Prism.languages[grammarName];
        if (!grammar) {
            return [];
        }

        // Tokenize the whole block text
        var texts = block.getTexts();
        var blockText = texts.map(function (t) {
            return t.text;
        }).join('\n');
        var tokens = Prism.tokenize(blockText, grammar);

        // The list of decorations to return
        var decorations = [];
        var textStart = 0;
        var textEnd = 0;

        texts.forEach(function (text) {
            textEnd = textStart + text.text.length;

            var offset = 0;

            function processToken(token, accu) {
                accu = accu || '';

                if (typeof token === 'string') {
                    if (accu) {
                        var decoration = createDecoration({
                            text: text,
                            textStart: textStart,
                            textEnd: textEnd,
                            start: offset,
                            end: offset + token.length,
                            className: 'prism-token token ' + accu
                        });
                        if (decoration) {
                            decorations.push(decoration);
                        }
                    }
                    offset += token.length;
                } else {
                    accu = accu + ' ' + token.type + ' ' + (token.alias || '');

                    if (typeof token.content === 'string') {
                        var _decoration = createDecoration({
                            text: text,
                            textStart: textStart,
                            textEnd: textEnd,
                            start: offset,
                            end: offset + token.content.length,
                            className: 'prism-token token ' + accu
                        });
                        if (_decoration) {
                            decorations.push(_decoration);
                        }

                        offset += token.content.length;
                    }

                    // When using token.content instead of token.matchedStr, token can be deep
                    else {
                            for (var i = 0; i < token.content.length; i++) {
                                processToken(token.content[i], accu);
                            }
                        }
                }
            }

            tokens.forEach(processToken);
            textStart = textEnd + 1; // account for added `\n`
        });

        return decorations;
    }

    return {
        schema: {
            rules: [{
                match: opts.onlyIn,
                decorate: decorate
            }, {
                match: function match(object) {
                    return object.kind === 'mark' && object.type === MARK_TYPE;
                },
                render: opts.renderToken
            }]
        }
    };
}

/**
 * Return a decoration range for the given text.
 */
function createDecoration(_ref) {
    var text = _ref.text,
        textStart = _ref.textStart,
        textEnd = _ref.textEnd,
        start = _ref.start,
        end = _ref.end,
        className = _ref.className;

    if (start >= textEnd || end <= textStart) {
        // Ignore, the token is not in the text
        return;
    }

    // Shrink to this text boundaries
    start = Math.max(start, textStart);
    end = Math.min(end, textEnd);

    // Now shift offsets to be relative to this text
    start = start - textStart;
    end = end - textStart;

    return {
        anchorKey: text.key,
        anchorOffset: start,
        focusKey: text.key,
        focusOffset: end,
        marks: [{ type: 'prism-token', data: { className: className } }]
    };
}

module.exports = PrismPlugin;